"""
This package contains an abstraction for a git repository.
"""
from base64 import b64encode
from datetime import timedelta
from enum import Enum
from json.decoder import JSONDecodeError
import time
from typing import Callable
from typing import Dict
from typing import Optional

from backoff import on_exception, expo
from requests.auth import AuthBase
from requests.auth import HTTPBasicAuth
import requests

from IGitt.Utils import Cache


HEADERS = {'User-Agent': 'IGitt'}


class IGittObject:
    """
    Any IGitt interface should inherit from this and any IGitt object shall
    have those methods.
    """

    @property
    def hoster(self):
        """
        The hosting service of the object, e.g. 'gitlab' or 'github'.
        """
        raise NotImplementedError

    @property
    def url(self):
        """
        Returns API url.
        """
        raise NotImplementedError

    @property
    def web_url(self):
        """
        Returns the web link.
        """
        raise NotImplementedError

    def __eq__(self, other):
        """
        Wether or not self is equal to another object :)
        """
        return hash(self) == hash(other)

    def __hash__(self):
        """
        A unique hash.
        """
        return hash(self.url)


class Token:
    """
    Base class for different types of tokens used for different methods of
    authentications.
    """
    @property
    def headers(self):
        """
        The Authorization headers.
        """
        raise NotImplementedError

    @property
    def value(self):
        """
        Token value
        """
        raise NotImplementedError

    @property
    def parameter(self):
        """
        Parameter to be used for authentication
        """
        raise NotImplementedError

    @property
    def auth(self):
        """
        A AuthBase instance that can be readily used to configure any kind of
        authentication easily with requests library.
        """
        raise NotImplementedError


class BasicAuthorizationToken(Token):
    """
    Basic HTTP Authorization using username and password.
    """
    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password
        self._encoded = None

    @property
    def value(self):
        if not self._encoded:
            self._encoded = b64encode(bytes('{}:{}'.format(
                self.username, self.password), 'utf-8'))
        return self._encoded

    @property
    def headers(self):
        """
        Addtional headers are not required as HTTPBasicAuth is being used.
        """
        return {}

    @property
    def parameter(self):
        """
        Basic HTTP Authentication only refers to use of `Authorization` Header.
        """
        return {}

    @property
    def auth(self):
        return HTTPBasicAuth(self.username, self.password)


def is_client_error_or_unmodified(exception):
    """
    Returns true if the request responded with a client error.
    """
    return (400 <= exception.args[1] < 500) or (exception.args[1] == 304)


def parse_response(response: requests.Response):
    """
    Parses the response object into JSON and link headers and returns them.
    """
    try:
        return response.json(), response.links
    except JSONDecodeError:
        # if the response body is pure text, for e.g. a git diff.
        return response.text, response.links


@on_exception(expo, ConnectionError, max_tries=8)
@on_exception(expo,
              RuntimeError,
              max_tries=3,
              giveup=is_client_error_or_unmodified)
def get_response(method: Callable,
                 url: str,
                 auth: AuthBase,
                 json: Optional[Dict]=frozenset()):
    """
    Sends a request and returns the response. Also checks the response for
    errors, and keeps retrying unless it's a HTTP Client Error.
    """
    if method.__name__.lower() != 'get':
        resp = method(url, auth=auth, json=dict(json or {}))
        if resp.status_code >= 300 and resp.status_code != 304:
            raise RuntimeError(resp.text, resp.status_code)
        return parse_response(resp)

    # cache only GET requests
    cached_resp, headers = Cache.get(url), {}
    if cached_resp:
        if cached_resp['fromWebhook']:
            headers['If-Modified-Since'] = cached_resp.get('lastFetched')
        else:
            headers['If-None-Match'] = cached_resp.get('entityTag')
    resp = method(url, auth=auth, json=dict(json or {}), headers=headers)
    if resp.status_code == 304 and cached_resp:
        return cached_resp.get('data'), cached_resp.get('links')
    elif resp.status_code >= 300:
        raise RuntimeError(resp.text, resp.status_code)

    data, links = parse_response(resp)
    # update entry in cache
    Cache.set(url, {
        'entityTag': resp.headers.get('ETag'),
        'data': data,
        'links': links
    })
    return data, links


def _fetch(url: str, req_type: str, token: Token, data: Optional[dict]=None,
           query_params: Optional[dict]=None, headers: Optional[dict]=None):
    """
    Fetch all the contents by following the ``Link`` header.

    :param url:
        The URL to query.
    :param req_type:
        The request type. Get, Post, Patch and Delete.
    :param token:
        The Token object to be used for authentication.
    :param data:
        The data to post. Used for PATCH and POST methods only.
    :param query_params:
        Any additional query parameters that should be sent with the request.
    :param headers:
        Any additional headers that should be sent with request.
    :return:
        A dictionary or a list of dictionaries if the response contains
        multiple items (usually in case of pagination) or a string in case of
        other format received (e.g. when fetching a git patch or diff) and the
        corresponding HTTP status code.
    """
    data_container = []
    session = requests.Session()
    session.headers.update({**dict(headers or {}), **HEADERS, **token.headers})
    session.params.update({**dict(query_params or {}), **token.parameter})
    req_methods = {
        'get': session.get,
        'post': session.post,
        'put': session.put,
        'patch': session.patch,
        'delete': session.delete
    }
    method = req_methods[req_type.lower()]
    resp, links = get_response(method, url, token.auth, json=data)

    # if the response body is pure text
    if isinstance(resp, str):
        # if the response body is empty, for e.g. in case of a DELETE request
        if resp == '':
            return []
        return resp

    while True:
        if isinstance(resp, dict):
            if 'items' in resp:
                # if response is a dict with `items` key, i.e. a list of items
                data_container.extend(resp['items'])
            else:
                # if response is a single item
                return resp
        elif isinstance(resp, list):
            # if response is a list of items
            data_container.extend(resp)
        if not links.get('next', False):
            return data_container
        resp, links = get_response(
            method, links.get('next')['url'], token.auth, json=data)


def get(token: Token, url: str, params: Optional[dict]=None,
        headers: Optional[dict]=None):
    """
    Queries the given URL for data.

    :param token: A token.
    :param url: The URL to access.
    :param params: The query params to be sent.
    :param headers: The request headers to be sent.
    :return:
        A dictionary or a list of dictionary if the response contains multiple
        items (usually in case of pagination) and the HTTP status code.
    :raises RunTimeError:
        If the response indicates any problem.
    """
    return _fetch(url, 'get', token,
                  query_params={**dict(params or {}), 'per_page': 100},
                  headers=headers)


def post(token: Token, url: str, data: dict, headers: Optional[dict]=None):
    """
    Posts the given data to the given URL.

    :param token: A token.
    :param url: The URL to access.
    :param data: The data to post.
    :param headers: The request headers to be sent.
    :return:
        A dictionary or a list of dictionary if the response contains multiple
        items (usually in case of pagination) and the HTTP status code.
    :raises RunTimeError:
        If the response indicates any problem.
    """
    return _fetch(url, 'post', token, data, headers=headers)


def put(token: Token, url: str, data: dict, headers: Optional[dict]=None):
    """
    Puts the given data to the given URL.

    :param token: A token.
    :param url: The URL to access.
    :param data: The data to put.
    :param headers: The request headers to be sent.
    :return:
        A dictionary or a list of dictionary if the response contains multiple
        items (usually in case of pagination) and the HTTP status code.
    :raises RunTimeError:
        If the response indicates any problem.
    """
    return _fetch(url, 'put', token, data, headers=headers)


def patch(token: Token, url: str, data: dict, headers: Optional[dict]=None):
    """
    Patches the given data to the given URL.

    :param token: A token.
    :param url: The URL to access.
    :param data: The data to patch.
    :param headers: The request headers to be sent.
    :return:
        A dictionary or a list of dictionary if the response contains multiple
        items (usually in case of pagination) and the HTTP status code.
    :raises RunTimeError:
        If the response indicates any problem.
    """
    return _fetch(url, 'patch', token, data, headers=headers)


def delete(token:Token, url: str, data: Optional[dict]=None,
           headers: Optional[dict]=None, params: Optional[dict]=None):
    """
    Sends a delete request to the given URL.

    :param token: A token.
    :param url: The URL to access.
    :param params: The query params to be sent.
    :param headers: The request headers to be sent.
    :raises RuntimeError: If the response indicates any problem.
    """
    _fetch(url, 'delete', token, data, query_params=params, headers=headers)


async def lazy_get(url: str,
                   callback: Callable,
                   headers: Optional[dict]=None,
                   timeout: Optional[timedelta]=timedelta(seconds=120),
                   interval: Optional[timedelta]=timedelta(seconds=10)):
    """
    Queries GitHub on the given URL for data, waiting while it
    returns HTTP 202.

    :param url: The full URL to query.
    :param callback:
        The function to callback with data after data is obtained.
        An empty dictionary is sent if nothing is returned by the API.
    :param timeout: datetime.timedelta object with time to keep re-trying.
    :param interval:
        datetime.timedelta object with time to keep in between tries.
    :param headers: The request headers to be sent.
    """
    response = requests.get(url, headers=headers, timeout=3000)

    # Wait and re-request to allow github to process query
    while response.status_code == 202 and timeout.total_seconds() > 0:
        time.sleep(interval.total_seconds())
        timeout -= interval
        response = requests.get(url, headers=headers, timeout=3000)

    await callback(response.json())


class AccessLevel(Enum):
    """
    Different access levels for users.
    """
    NONE = 0  # in case of private repositories
    CAN_VIEW = 10
    CAN_READ = 20
    CAN_WRITE = 30
    ADMIN = 40
    OWNER = 50

class MergeRequestStates(Enum):
    """
    This class depicts the merge request states that can are present in any
    hosting service providers like GitHub or GitLab.
    """
    OPEN = 'open'
    CLOSED = 'closed'
    MERGED = 'merged'

class IssueStates(Enum):
    """
    This class depicts the issue states that can are present in any hosting
    service providers like GitHub or GitLab.
    """
    def __str__(self):
        """
        Make behaviour of object as similar to a string as possible.
        """
        return str(self.value)

    OPEN = 'open'
    CLOSED = 'closed'

class MilestoneStates(Enum):
    """
    This class depicts the milestone states that can are present in any hosting
    service providers like GitHub or GitLab.
    """
    def __str__(self):
        """
        Make behaviour of object as similar to a string as possible.
        """
        return str(self.value)

    OPEN = 'open'
    CLOSED = 'closed'



"""
This package contains the GitHub implementations of the interfaces in
server.git.Interfaces.
"""
from datetime import datetime
from typing import Optional
import os
import logging

from requests_oauthlib import OAuth2
import jwt

from IGitt.Interfaces import Token, get, post
from IGitt.Utils import CachedDataMixin


GH_INSTANCE_URL = os.environ.get('GH_INSTANCE_URL', 'https://github.com')
if not GH_INSTANCE_URL.startswith('http'):  # dont cover cause it'll be removed
    GH_INSTANCE_URL = 'https://' + GH_INSTANCE_URL
    logging.warning('Include the protocol in GH_INSTANCE_URL! Omitting it has '
                    'been deprecated.')
BASE_URL = GH_INSTANCE_URL.replace('github.com', 'api.github.com')


class GitHubMixin(CachedDataMixin):
    """
    Base object for things that are on GitHub.
    """

    def _get_data(self):
        return get(self._token, self.url)

    @staticmethod
    def absolute_url(url):
        """
        Makes a URL like ``/repo/coala/coala`` absolute.
        """
        return BASE_URL + url

    @property
    def hoster(self):
        """
        Returns `github`.
        """
        return 'github'

    @property
    def url(self):
        """
        Returns github API url.
        """
        return self.absolute_url(self._url)

    @property
    def web_url(self):
        """
        Returns the web link for GitHub.
        """
        return self.data['html_url']

    def __repr__(self): # dont cover
        return '<{} object(url={}) at {}>'.format(self.__class__.__name__,
                                                  self.url,
                                                  hex(id(self)))


class GitHubToken(Token):
    """
    Object representation of oauth tokens.
    """

    def __init__(self, token):
        self._token = token

    @property
    def headers(self):
        """
        GitHub Access token does not require any special headers.
        """
        return {}

    @property
    def parameter(self):
        """
        No additional query parameters are used with GitHub token.
        """
        return {}

    @property
    def value(self):
        return self._token

    @property
    def auth(self):
        return OAuth2(token={'access_token': self._token,
                             'token_type': 'bearer'})


class GitHubJsonWebToken(Token):
    """
    Object representation of JSON Web Token.
    """
    def __init__(self, private_key: str, app_id: int):
        self._key = private_key.strip()
        self._app_id = app_id
        self._payload = None
        self._jwt_token = None

    @property
    def payload(self):
        """
        Returns the payload to be sent for JWT encoding.
        """
        if not self._payload:
            self._payload = {
                # issued at time
                'iat': int(datetime.now().timestamp()),
                # JWT expiration time (10 minute maximum), minus 5 seconds just
                # to be sure and cover up the request time
                'exp': int(datetime.now().timestamp() + (10 * 60) - 5),
                # GitHub App's identifier
                'iss': self._app_id
            }
        return self._payload

    # testing over recorded requests is unadvisable as it is dependent on the
    # time of execution of tests
    @property
    def is_expired(self):  # dont cover
        """
        Returns True if the JWT has expired.
        """
        return self.payload['exp'] < datetime.now().timestamp()

    @property
    def headers(self):
        return {'Authorization': 'Bearer {}'.format(self.value),
                'Accept': 'application/vnd.github.machine-man-preview+json'}

    @property
    def parameter(self):
        """
        GitHub's JSON Web Token can only be authenticated via the
        ``Authorization`` header and so, all the nested requests have to be made
        in only that way.
        """
        return {}

    @property
    def value(self):
        if not self._jwt_token or self.is_expired:
            self._jwt_token = jwt.encode(self.payload, self._key, 'RS256')
        return self._jwt_token.decode('utf-8')

    @property
    def auth(self):
        """
        OAuth 2.0 JWT Bearer Token Flow is not supported by oauthlib yet.

        Reference: https://github.com/oauthlib/oauthlib/issues/50
        """
        return None


class GitHubInstallationToken(Token):
    """
    Object representation of GitHub Installation Token.
    """
    def __init__(self,
                 installation_id: int,
                 jwt_token: GitHubJsonWebToken,
                 token: Optional[str]=None,
                 expiry: Optional[datetime]=None):
        self._jwt = jwt_token
        self._expiry = expiry
        self._token = token
        self._id = installation_id

    @property
    def jwt(self):
        """
        Retrieves the JWT being used.
        """
        return self._jwt

    @property
    def headers(self):
        return {'Authorization': 'token {}'.format(self.value),
                'Accept': 'application/vnd.github.machine-man-preview+json'}

    # testing over recorded requests is unadvisable as it is dependent on the
    # time of execution of tests
    @property
    def is_expired(self):  # dont cover
        """
        Returns true if the token has expired.
        """
        if not self._expiry:
            return True
        return datetime.utcnow() > self._expiry

    def _get_new_token(self):
        data = post(self._jwt,
                    BASE_URL+'/installations/{}/access_tokens'.format(self._id),
                    {})
        return data['token'], datetime.strptime(data['expires_at'],
                                                '%Y-%m-%dT%H:%M:%SZ')

    @property
    def value(self):
        if self.is_expired or not self._token:
            self._token, self._expiry = self._get_new_token()
        return self._token

    @property
    def parameter(self):
        """
        GitHub Installation Token can only be authenticated via the
        ``Authorization`` header and so, all the nested requests have to be
        made in only that way.
        """
        return {}

    @property
    def auth(self):
        """
        OAuth 2.0 JWT Bearer Token Flow is not supported by oauthlib yet.

        Reference: https://github.com/oauthlib/oauthlib/issues/50
        """
        return None



import pytest


def pytest_addoption(parser):
    parser.addoption(
        '--record-cassettes',
        action='store',
        default='once',
        help=('Modify vcrpy cassette recording mode. Allowed options are '
              'once (record cassettes once), '
              'new_episodes (records only new cassettes and replay old ones), '
              'none (only replay previous cassettes) and '
              'all (record everything, never replay recorded interactions).'))


@pytest.fixture(scope='class')
def vcrpy_record_mode(request):
    request.cls.vcrpy_record_mode = request.config.getoption(
        '--record-cassettes')



"""
Contains a representation of GitHub users.
"""
from typing import Optional

from IGitt.GitHub import GitHubMixin
from IGitt.GitHub import GitHubToken
from IGitt.GitHub import GitHubInstallationToken
from IGitt.Interfaces import get
from IGitt.Interfaces.User import User


PREVIEW_HEADER = {'Accept': 'application/vnd.github.machine-man-preview+json'}


class GitHubUser(GitHubMixin, User):
    """
    A GitHub user, e.g. sils :)
    """
    def __init__(self, token: GitHubToken, username: Optional[str]=None):
        """
        Creates a GitHubUser with the given credentials.

        :param token:       The Oauth token.
        :param username:    The username of github user.
                            Takes None, if it's the associated user's token.
        """
        self._token = token
        self._url = '/users/' + username if username else '/user'
        self._username = username

    @property
    def username(self) -> str:
        """
        Retrieves the login for the user.
        """
        return self._username or self.data['login']

    @property
    def identifier(self) -> int:
        """
        Gets a unique id for the user that never changes.
        """
        return self.data['id']

    def installed_repositories(self, installation_id: int):
        """
        List repositories that are accessible to the authenticated user for an
        installation.
        """
        # Don't move to module code, causes circular dependencies
        from IGitt.GitHub.GitHubRepository import GitHubRepository

        repos = get(self._token,
                    self.absolute_url(
                        '/user/installations/{}/repositories'.format(
                            installation_id
                        )
                    ),
                    headers=PREVIEW_HEADER)['repositories']
        return {GitHubRepository.from_data(repo, self._token, repo['id'])
                for repo in repos}

    def get_installations(self, jwt):
        """
        Gets the installations this user has access to.

        :param jwt: The GitHubJsonWebToken required to fetch data.
        """
        # Don't move to module code, causes circular dependencies
        from IGitt.GitHub.GitHubInstallation import GitHubInstallation

        resp = get(self._token, self.absolute_url('/user/installations'),
                   headers=PREVIEW_HEADER)
        return {
            GitHubInstallation.from_data(
                i, GitHubInstallationToken(i['id'], jwt), i['id'])
            for i in resp['installations']
        }



"""
This module contains the Issue abstraction class which provides properties and
actions related to issues and bug reports.
"""
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional
from typing import Set

from IGitt.Interfaces import IGittObject
from IGitt.Interfaces.Comment import Comment
from IGitt.Interfaces.Repository import Repository
from IGitt.Interfaces.User import User
from IGitt.Interfaces.Milestone import Milestone


class Issue(IGittObject):
    """
    Represents an issue on GitHub or GitLab or a bug report on bugzilla or so.
    """

    @property
    def number(self) -> int:
        """
        Returns the issue "number" or id.
        """
        raise NotImplementedError

    @property
    def repository(self) -> Repository:
        """
        Returns the repository this issue is linked with.
        """
        raise NotImplementedError

    @property
    def title(self) -> str:
        """
        Retrieves the title of the issue.
        """
        raise NotImplementedError

    @title.setter
    def title(self, new_title):
        """
        Sets the title of the issue.

        :param new_title: The new title.
        """
        raise NotImplementedError

    @property
    def description(self) -> str:
        """
        Retrieves the main description of the issue.
        """
        raise NotImplementedError

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the issue.

        :param new_description: The new description.
        """
        raise NotImplementedError

    @property
    def author(self) -> User:
        """
        Retrieves the author of the comment wrapped in a User object.
        """
        raise NotImplementedError

    @property
    def assignees(self) -> Set[User]:
        """
        Retrieves a set of usernames of assignees.
        """
        raise NotImplementedError

    @assignees.setter
    def assignees(self, value: Set[User]):
        """
        Setter for ssignees.
        """
        raise NotImplementedError

    def assign(self, *usernames: List[User]):
        """
        Sets a given users as assignee.
        """
        raise NotImplementedError

    def unassign(self, *usernames: List[User]):
        """
        Unassigns given users from issue.
        """
        raise NotImplementedError

    @property
    def available_assignees(self) -> Set[User]:
        """
        Compiles a set of Users that are available for assigning this issue.

        :return: A set of User objects.
        """
        raise NotImplementedError

    @property
    def labels(self) -> Set[str]:
        """
        Retrieves the set of labels the issue is currently tagged with.

        :return: The set of labels.
        """
        raise NotImplementedError

    @labels.setter
    def labels(self, value: Set[str]):
        """
        Tags the issue with the given labels. For examples see documentation
        of the labels read function.

        Labels are added and removed as necessary on remote.

        :param value: The new set of labels.
        """
        raise NotImplementedError

    @property
    def available_labels(self) -> Set[str]:
        """
        Compiles a set of labels that are available for labelling this issue.

        :return: A set of label captions.
        """
        raise NotImplementedError

    @property
    def comments(self) -> List[Comment]:
        """
        Retrieves a list of comments which are on the issue excliding the
        description.

        :return: A list of Comment objects.
        """
        raise NotImplementedError

    def add_comment(self, body) -> Comment:
        """
        Adds a comment to the issue.

        :param body: The content of the comment.
        :return: The newly created comment.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def close(self):
        """
        Closes the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def reopen(self):
        """
        Reopens the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def delete(self):
        """
        Deletes the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @property
    def state(self) -> str:
        """
        Get's the state of the issue.

        :return: Either 'open' or 'closed'.
        """
        raise NotImplementedError

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was created.
        """
        raise NotImplementedError

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was updated the last time.
        """
        raise NotImplementedError

    @property
    def reactions(self) -> List[str]:
        """
        Retrieves the reactions / award emojis applied on the issue.
        """
        raise NotImplementedError

    @staticmethod
    def create(token, repository, title, body='', issue_type=None):
        """
        Create a new issue in repository.
        """
        raise NotImplementedError

    @property
    def mrs_closed_by(self) -> Set:
        """
        Returns the merge requests that close this issue.
        """
        raise NotImplementedError

    @property
    def milestone(self) -> Milestone:
        """
        Retrieves the milestone.
        """
        raise NotImplementedError

    @milestone.setter
    def milestone(self, new_milestone) -> Milestone:
        """
        Setter for the Milestone.
        """
        raise NotImplementedError

    @property
    def time_estimate(self) -> timedelta:
        """
        Retrieves the time_estimate in seconds.
        Writes the time_estimate into
        the seconds property of an timedelta object.
        """
        raise NotImplementedError

    @time_estimate.setter
    def time_estimate(self, new_time_estimate: timedelta):
        """
        Setter for the time_estimate.
        Allows any time unit of the timedelta object.
        """
        raise NotImplementedError

    @property
    def total_time_spent(self) -> timedelta:
        """
        Retrieves the total_time_spent in seconds.
        Writes the total_time_spent into the
        seconds property of an timedelta object.
        """
        raise NotImplementedError

    @property
    def weight(self) -> Optional[int]:
        """
        Retrieves the weight associated with the current issue.
        """
        raise NotImplementedError

    @weight.setter
    def weight(self, value: int):
        """
        Updates the weight associated with the current issue.
        """
        raise NotImplementedError

    @total_time_spent.setter
    def total_time_spent(self, absolute_time_spent: timedelta):
        """
        Writes the value of absolute_time_spent into total_time_spent.
        Can't be less than 0.
        Allows any time unit of the timedelta object.
        Allows total_time_spent to be reset to 0 by passing a None or 0.
        """
        raise NotImplementedError

    def add_to_total_time_spent(self, relative_time_spent: timedelta):
        """
        Adds the value of relative_time_spent to total_time_spent.
        Allows for positive and negative values.
        Allows any time unit of the timedelta object.
        Does nothing when passing None or 0.
        """
        raise NotImplementedError



"""
Contains the Notification base class.
"""
from enum import Enum
from typing import Union

from IGitt.Interfaces import IGittObject
from IGitt.Interfaces import Token
from IGitt.Interfaces.Commit import Commit
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces.MergeRequest import MergeRequest
from IGitt.Interfaces.Repository import Repository


class Reason(Enum):
    """
    The enum representing the reason/action of a notification.
    """
    ASSIGNED = 'assigned'
    MENTIONED = 'mentioned'
    BUILD_FAILED = 'build_failed'
    MARKED = 'marked'
    APPROVAL_REQUIRED = 'approval_required'
    AUTHORED = 'authored'
    COMMENTED = 'commented'
    INVITED = 'invited'
    MANUAL = 'manual'
    STATE_CHANGED = 'state_changed'
    SUBSCRIBED = 'subscribed'


class Notification(IGittObject):
    """
    Represents a notification/todo on GitHub or GitLab.
    """
    @staticmethod
    def fetch_all(token: Token):
        """
        Returns the list of notifications for the user bearing the token.
        """
        raise NotImplementedError

    @property
    def reason(self) -> Reason:
        """
        Returns the reason for notification.
        """
        raise NotImplementedError

    @property
    def subject_type(self) -> type:
        """
        Returns the type of the subject of the notification.
        """
        raise NotImplementedError

    @property
    def subject(self) -> Union[Commit, Issue, MergeRequest]:
        """
        Returns the subject of notification.
        """
        raise NotImplementedError

    @property
    def pending(self) -> bool:
        """
        Returns True if the notification is pending/not read.
        """
        raise NotImplementedError

    @property
    def repository(self) -> Repository:
        """
        Returns the repository this notification is related to.
        """
        raise NotImplementedError

    @property
    def identifier(self) -> str:
        """
        Returns the id of the notification.
        """
        raise NotImplementedError

    def mark_done(self):
        """
        Marks the notification as done/read.
        """
        raise NotImplementedError

    def unsubscribe(self):
        """
        Unsubscribe from this subject.
        """
        raise NotImplementedError



"""
This module contains the Issue abstraction class which provides properties and
actions related to issues and bug reports.
"""
import re
from functools import lru_cache
from typing import Set
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

from IGitt.GitLab import GitLabPrivateToken
from IGitt.GitLab import GitLabOAuthToken
from IGitt.GitLab import GL_INSTANCE_URL
from IGitt.GitLab import GitLabMixin
from IGitt.GitLab.GitLabUser import GitLabUser
from IGitt.GitLab.GitLabIssue import GitLabIssue
from IGitt.Interfaces import get
from IGitt.Interfaces import post
from IGitt.Interfaces import AccessLevel
from IGitt.Interfaces.Organization import Organization
from IGitt.Interfaces.Repository import Repository


class GitLabOrganization(GitLabMixin, Organization):
    """
    Represents an organization on GitLab.
    """
    @property
    def web_url(self):
        return '{}/{}'.format(GL_INSTANCE_URL, self.name)

    def __init__(self, token, name):
        """
        :param name: Name of the org.
        """
        self._token = token
        self._name = name
        self._url = '/groups/{name}'.format(name=quote_plus(name))
        self._is_user = None

    @property
    def identifier(self) -> int:
        """
        Returns the identifier of the organization.
        """
        return self.data['id']

    @property
    def description(self) -> str:
        """
        Returns the description of this organization.
        """
        return self.data['description']

    @lru_cache()
    def raw_members(self) -> list:
        """
        Gets all members including the ones of groups around subgroups as a
        list of dicts from the JSON responses.
        """
        members = list(get(self._token, self.url + '/members'))
        if '/' in self.name:
            members.extend(GitLabOrganization(
                self._token,
                self.name.rsplit('/', maxsplit=1)[0]).raw_members())

        return members

    @property
    def billable_users(self) -> int:
        """
        Number of paying/registered users on the organization.
        """
        try:
            # If the org is a user, this'll throw RuntimeError
            users = {user['username'] for user in get(self._token,
                                                      self.url + '/members')}

            for group in get(self._token, self.absolute_url('/groups')):
                gname = group['full_path']
                if (gname in self.name or self.name in gname) \
                        and self.name != gname:
                    users |= {
                        user['username']
                        for user in get(
                            self._token,
                            self.absolute_url('/groups/{name}/members'.format(
                                name=quote_plus(gname)
                            ))
                        )
                    }

            return len(users)
        except RuntimeError:
            return 1

    def _members(self, access_level):
        try:
            return {
                GitLabUser.from_data(user, self._token, user['id'])
                for user in self.raw_members()
                if user['access_level'] >= access_level
            }
        except RuntimeError:
            return {GitLabUser.from_data({'username': self.name},
                                         self._token, identifier=None)}

    @property
    def owners(self) -> Set[GitLabUser]:
        """
        Returns the user handles of all owner users.
        """
        return self._members(AccessLevel.OWNER.value)

    @property
    def masters(self) -> Set[GitLabUser]:
        """
        Returns the user handles of all master users.
        """
        return self._members(AccessLevel.ADMIN.value)

    @property
    def name(self) -> str:
        """
        The name of the organization.
        """
        return self._name

    @property
    @lru_cache(None)
    def suborgs(self) -> Set[Organization]:
        """
        Returns the sub-organizations within this organization, recursively.
        """
        result = set()
        for suborg_data in get(self._token, self.url + '/subgroups'):
            suborg = GitLabOrganization.from_data(
                suborg_data, self._token, suborg_data['full_path'])
            result |= {suborg} | suborg.suborgs
        return result

    @property
    @lru_cache(None)
    def repositories(self) -> Set[Repository]:
        """
        Returns the list of repositories contained in this organization
        including subgroup repositories, recursively.
        """
        from IGitt.GitLab.GitLabRepository import GitLabRepository

        return {GitLabRepository.from_data(repo, self._token, repo['id'])
                for repo in get(self._token, self.url + '/projects')
               }.union({
                   repo for org in self.suborgs for repo in org.repositories
               })

    def filter_issues(self,
                      state: Optional[str]=None,
                      label: Optional[str]=None,
                      assignee: Optional[str]=None
                     ) -> Set[GitLabIssue]:
        """
        Filters the issues in the organization based on properties

        :param state: 'opened' or 'closed' or 'all'
        :param label: Label of the issue
        :param assignee: username of issue assignee
        :return: Set of GitLabIssue objects
        """
        params = dict()
        if state:
            params['state'] = state
        if label:
            params['labels'] = label
        if assignee:
            params['assignee_id'] = GitLabUser(self._token,
                                               assignee).identifier
        url = re.compile(r'https://(?:[^/]+)/(.+)/issues/(\d+)')
        return {GitLabIssue.from_data(issue, self._token,
                                      url.match(issue['web_url']).group(0),
                                      issue['iid'])
                for issue in get(self._token, self.url + '/issues', params)}

    @property
    def issues(self) -> Set[GitLabIssue]:
        """
        Returns the list of issue objects in this organization.
        """
        return self.filter_issues(state='opened')

    @staticmethod
    def create(token: Union[GitLabOAuthToken, GitLabPrivateToken],
               name: str,
               path: str,
               parent_id: Optional[int]=None,
               description: Optional[str]=None,
               visibility: str='private',
               lfs_enabled: bool=False,
               request_access_enabled: bool=False) -> Organization:
        """
        Creates a new organization from the given parameters.

        :param token:
            The credentials to be used for authorization.
        :param name:
            The name of the organization.
        :param path:
            The path of the organization.
        :param parent_id:
            The parent organization id to create nested organization.
        :param description:
            The description of the organization.
        :param visibility:
            Controls the visibility of the organization. Can be either
            'private', 'public' or 'internal'.
        :param lfs_enabled:
            Enables Git Large File System for projects in this organization.
        :param request_access_enabled:
            Allow users to request member access on the organization.
        """
        url = '/groups'
        org = post(
            token,
            GitLabOrganization.absolute_url(url),
            {'name': name,
             'path': path,
             'description': description,
             'visibility': visibility,
             'lfs_enabled': lfs_enabled,
             'request_access_enabled': request_access_enabled,
             'parent_id': parent_id}
        )
        return GitLabOrganization.from_data(org, token, org['name'])



"""
Contains the GitHub Team implementation.
"""
from typing import Set

from IGitt.GitHub import GH_INSTANCE_URL
from IGitt.GitHub import GitHubMixin
from IGitt.GitHub.GitHubUser import GitHubUser
from IGitt.GitHub.GitHubOrganization import GitHubOrganization
from IGitt.Interfaces.Team import Team
from IGitt.Interfaces import get


class GitHubTeam(GitHubMixin, Team):
    """
    Represents a team on GitHub.
    """

    def __init__(self, token, team_id):
        """
        :param team_id: The id of the team.
        """
        self._token = token
        self._id = team_id
        self._url = '/teams/{team_id}'.format(team_id=team_id)


    @property
    def name(self) -> str:
        """
        Name of the team.
        """
        return self.data['name']

    @property
    def id(self) -> int:
        """
        Retrieves the id of the team.
        """
        return self._id

    @property
    def description(self) -> str:
        """
        Returns the description of this team.
        """
        return self.data['description']

    @property
    def web_url(self):
        return '{}/orgs/{}/teams/{}'.format(GH_INSTANCE_URL,
                                            self.get_organization.name,
                                            self.name)

    @property
    def members(self) -> Set[GitHubUser]:
        """
        Returns the user handles of all members of this team.
        """
        return {
            GitHubUser.from_data(user, self._token, user['login'])
            for user in get(
                self._token, self.url + '/members'
            )
        }

    def is_member(self, username):
        """
        Checks if given username is member of this team.
        """
        usernames = [user.username for user in self.members]
        return True if username in usernames else False

    @property
    def get_organization(self):
        """
        Returns parent organization.
        """
        return GitHubOrganization.from_data(self.data['organization'],
                                            self._token,
                                            self.data['organization']['login'])



"""
Contains an object representation of a reaction / award emoji on GitLab.
"""
from typing import Union

from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import GitLabOAuthToken
from IGitt.GitLab import GitLabPrivateToken
from IGitt.GitLab.GitLabUser import GitLabUser
from IGitt.Interfaces.Comment import Comment
from IGitt.Interfaces.MergeRequest import MergeRequest
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces.Reaction import Reaction


class GitLabReaction(GitLabMixin, Reaction):
    """
    A GitLab Reaction or Award Emoji, e.g. heart.
    """
    def __init__(self,
                 token: Union[GitLabOAuthToken, GitLabPrivateToken],
                 related: Union[Issue, MergeRequest, Comment],
                 identifier: int):
        """
        Creates a GitLabReaction instance.

        :param token:       The authentication token.
        :param token:       The authentication token.
        :param related:     The object this reaction was added on,
                            i.e. an Issue, MergeRequest, Comment, etc.
        :param identifier:  The unique identifier of the reaction.
        """
        self._token = token
        self._related = related
        self._url = '{}/award_emoji/{}'.format(
            getattr(related, '_url'), identifier)
        self._identifier = identifier

    @property
    def name(self) -> str:
        """
        Retrieves the name of the reaction.
        """
        return self.data['name']

    @property
    def user(self) -> GitLabUser:
        """
        Retrieves the user who reacted with this reaction.
        """
        user = self.data['user']
        return GitLabUser.from_data(user, self._token, user['id'])



"""
Contains the abstraction for a commit in GitLab.
"""
from typing import Optional
from typing import Set
from typing import Union
from urllib.parse import quote_plus

from IGitt import ElementDoesntExistError
from IGitt.GitHub.GitHubCommit import get_diff_index
from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import GitLabOAuthToken, GitLabPrivateToken
from IGitt.GitLab.GitLabComment import GitLabComment
from IGitt.GitLab.GitLabRepository import GitLabRepository
from IGitt.GitLab.GitLabIssue import GitLabIssue
from IGitt.Interfaces import get, post
from IGitt.Interfaces.Comment import CommentType
from IGitt.Interfaces.Commit import Commit
from IGitt.Interfaces.CommitStatus import Status, CommitStatus

GL_STATE_TRANSLATION = {
    Status.RUNNING: 'running',
    Status.CANCELED: 'canceled',
    Status.ERROR: 'failed',
    Status.FAILED: 'failed',
    Status.PENDING: 'pending',
    Status.SUCCESS: 'success',
    Status.MANUAL: 'manual',
    Status.CREATED: 'created',
    Status.SKIPPED: 'skipped'
}

INV_GL_STATE_TRANSLATION = {val: key for key, val
                            in GL_STATE_TRANSLATION.items()}
GITLAB_KEYWORD_REGEX = {'fix': r'[Ff]ix(?:e[sd]|ing)?',
                        'close': r'[Cc]los(?:e[sd]?|ing)',
                        'resolve': r'[Rr]esolv(?:e[sd]?|ing)'}


class GitLabCommit(GitLabMixin, Commit):
    """
    Represents a commit on GitLab.
    """

    def __init__(self,
                 token: Union[GitLabOAuthToken, GitLabPrivateToken],
                 repository: str,
                 sha: Optional[str],
                 branch: Optional[str]=None):
        """
        Creates a new GitLabCommit object.

        :param token: A Token object to be used for authentication.
        :param repository: The full repository name.
        :param sha: The full commit SHA, if None given provide a branch.
        :param branch: A branch name if SHA is unavailable. Note that lazy
                       loading won't work in that case.
        """
        assert sha or branch, 'Either full SHA or branch name has to be given!'
        self._token = token
        self._repository = repository
        self._sha = sha
        self._branch = branch
        self._url = '/projects/{id}/repository/commits/{sha}'.format(
            id=quote_plus(repository), sha=sha if sha else branch)

    @property
    def message(self) -> str:
        """
        Returns the commit message.

        :return: Commit message as string.
        """
        return self.data['message']

    @property
    def sha(self):
        """
        Retrieves the SHA of the commit:

        >>> from os import environ
        >>> commit = GitLabCommit(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test', '674498'
        ... )
        >>> commit.sha
        '674498'

        :return: A string holding the SHA of the commit.
        """
        return self._sha if self._sha else self.data['id']

    @property
    def repository(self):
        """
        Retrieves the repository that holds this commit.

        >>> from os import environ
        >>> commit = GitLabCommit(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test', '3fc4b86'
        ... )
        >>> commit.repository.full_name
        'gitmate-test-user/test'

        :return: A usable Repository instance.
        """
        return GitLabRepository(self._token, self._repository)

    @property
    def parent(self):
        """
        Retrieves the parent commit. In case of a merge commit the first parent
        will be returned.

        >>> from os import environ
        >>> commit = GitLabCommit(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test', '3fc4b86'
        ... )
        >>> commit.parent.sha
        '674498fd415cfadc35c5eb28b8951e800f357c6f'

        :return: A Commit object.
        """
        return GitLabCommit(self._token, self._repository,
                            self.data['parent_ids'][0])

    def get_statuses(self) -> Set[CommitStatus]:
        """
        Retrieves the all commit statuses.

        :return: A (frozen)set of CommitStatus objects.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        # rebuild the url with full sha because gitlab doesn't work that way
        url = '/projects/{repo}/repository/commits/{sha}/statuses'.format(
            repo=quote_plus(self._repository), sha=self.sha)
        statuses = get(self._token, self.absolute_url(url))

        # Only the first of each context is the one we want
        result = set()
        contexts = set()
        for status in statuses:
            if status['name'] not in contexts:
                result.add(CommitStatus(
                    INV_GL_STATE_TRANSLATION[status['status']],
                    status['description'], status['name'],
                    status['target_url']))
                contexts.add(status['name'])

        return result

    @property
    def combined_status(self) -> Status:
        """
        Retrieves a combined status of all the commits.

        :return:
            Status.FAILED if any of the commits report as error or failure or
            canceled
            Status.PENDING if there are no statuses or a commit is pending or a
            test is running
            Status.SUCCESS if the latest status for all commits is success
        :raises AssertionError:
            If the status couldn't be matched with any of the possible outcomes
            Status.SUCCESS, Status.FAILED and Status.PENDING.
        """
        statuses = set(map(lambda status: status.status, self.get_statuses()))
        if (
                not len(statuses) or
                Status.PENDING in statuses or
                Status.RUNNING in statuses or
                Status.CREATED in statuses):
            return Status.PENDING
        if (
                Status.FAILED in statuses or
                Status.ERROR in statuses or
                Status.CANCELED in statuses):
            return Status.FAILED
        assert all(status in {Status.SUCCESS, Status.MANUAL}
                   for status in statuses)
        return Status.SUCCESS

    def set_status(self, status: CommitStatus):
        """
        Adds the given status to the commit.

        >>> from os import environ
        >>> commit = GitLabCommit(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test', '3fc4b86'
        ... )
        >>> status = CommitStatus(Status.FAILED, 'Theres a problem',
        ...                       'gitmate/test')
        >>> commit.set_status(status)
        >>> commit.get_statuses().pop().description
        'Theres a problem'

        If a status with the same context already exists, it will be bluntly
        overridden:

        >>> status.status = Status.SUCCESS
        >>> status.description = "Theres no problem"
        >>> commit.set_status(status)
        >>> len(commit.get_statuses())
        1
        >>> commit.get_statuses().pop().description
        'Theres no problem'

        :param status: The CommitStatus to set to this commit.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        data = {'state': GL_STATE_TRANSLATION[status.status],
                'target_url': status.url, 'description': status.description,
                'name': status.context}
        status_url = '/projects/{repo}/statuses/{sha}'.format(
            repo=quote_plus(self._repository), sha=self.sha)
        post(self._token, self.absolute_url(status_url), data)

    def get_patch_for_file(self, filename: str):
        r"""
        Retrieves the unified diff for the commit.

        >>> from os import environ
        >>> commit = GitLabCommit(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test', '3fc4b86'
        ... )
        >>> assert (commit.get_patch_for_file('README.md') ==
        ...         '--- a/README.md\n+++ b/README.md\n@@ -1,2 +1,4 @@\n '
        ...         '# test\n a test repo\n+\n+a tst pr\n')

        But only if it exists!

        >>> commit.get_patch_for_file('IDONTEXISTFILE')
        Traceback (most recent call last):
         ...
        IGitt.ElementDoesntExistError: The file does not exist.

        :param filename: The file to retrieve patch for.
        :return: A string containing the patch.
        :raises ElementDoesntExistError: If the given filename does not exist.
        """
        diff = get(self._token, self.url + '/diff')

        for patch in diff:
            if filename in (patch['new_path'], patch['old_path']):
                return patch['diff']

        raise ElementDoesntExistError('The file does not exist.')

    def comment(self, message: str, file: Optional[str]=None,
                line: Optional[int]=None,
                mr_number: Optional[int]=None) -> GitLabComment:
        """
        Places a comment on the commit.

        >>> from os import environ
        >>> commit = GitLabCommit(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test', '3fc4b86'
        ... )

        So this line places a comment on the bottom of the commit,
        not associated to any particular piece of code:

        >>> commit.comment("An issue is here!")

        However, we can also comment on a particular file and line, if that is
        included in the diff:

        >>> commit.comment("Here in line 4, there's a spelling mistake!",
        ...                'README.md', 4)

        If you supply the ``pr_number`` argument, the comment will appear in
        the review UI of that pull request:

        >>> commit.comment("Here in line 4, there's a spelling mistake!",
        ...                'README.md', 4, mr_number=7)

        Beat that! Of course, there's a lot of error handling. If you give the
        wrong file, the comment will appear below the commit with a note about
        the commit, file and line:

        >>> commit.comment("Oh, this'll end up below!!", 'READMENOT.md', 4)

        Also if the line isn't contained in the diff GitLab won't accept that
        and it'll also end up below - sorry!

        >>> commit.comment("Oh, this'll too end up below!!", 'README.md', 8)

        If you give a pull request, the comment will appear on the PR instead:

        >>> commit.comment("Oh, this'll too end up on the PR.",
        ...                'README.md', 8, mr_number=7)

        :param message: The body of the comment.
        :param file: The file to place the comment, relative to repo root.
        :param line: The line in the file in the comment or None.
        :param mr_number: The iid of a merge request if this should end up in
                          the discussions UI of the merge request.
        """
        data = {'note': message, 'line_type': 'new'}

        if file is not None and line is not None:
            try:
                patch = self.get_patch_for_file(file)
                index = get_diff_index(patch, line)
                if index:  # Else, fallback to comment below file
                    data['line'] = index
                    data['path'] = file
            except ElementDoesntExistError:
                pass  # Fallback to comment below the file

        if 'line' not in data:
            file_str = '' if file is None else ', file ' + file
            line_str = '' if line is None else ', line ' + str(line)
            data['note'] = ('Comment on ' + self.sha + file_str + line_str +
                            '.\n\n' + data['note'])

        # post a comment on commit
        if 'line' in data and 'path' in data or mr_number is None:
            url = '/projects/{id}/repository/commits/{sha}/comments'.format(
                id=quote_plus(self._repository), sha=self.sha)
            res = post(self._token, self.absolute_url(url), data)
            return

        # fallback to post the comment on relevant merge request
        if mr_number is not None:
            data['body'] = data['note']  # because gitlab is stupid
            url = '/projects/{id}/merge_requests/{mr_iid}/notes'.format(
                id=quote_plus(self._repository), mr_iid=mr_number)
            res = post(self._token, self.absolute_url(url), data)
            return GitLabComment.from_data(res, self._token, self._repository,
                                           mr_number, CommentType.MERGE_REQUEST,
                                           res['id'])

    @property
    def unified_diff(self):
        """
        Retrieves the unified diff for the commit excluding the diff index.
        """
        return '\n'.join(patch['diff']
                         for patch in get(self._token, self.url + '/diff')
                        )

    @property
    def closes_issues(self) -> Set[GitLabIssue]:
        """
        Returns a set of GitLabIssue objects which would be closed upon merging
        this pull request.
        """
        issues = self._get_closes_issues()
        return {GitLabIssue(self._token, repo_name, number)
                for number, repo_name in issues}

    @property
    def mentioned_issues(self) -> Set[GitLabIssue]:
        """
        Returns a set of GitLabIssue objects which are related to the merge
        request.
        """
        issues = self._get_mentioned_issues()
        return {GitLabIssue(self._token, repo_name, number)
                for number, repo_name in issues}

    @property
    def will_fix_issues(self) -> Set[GitLabIssue]:
        """
        Returns a set of GitLabIssue objects which would be fixed as stated in
        this commit message.
        """
        issues =  self.get_keywords_issues(GITLAB_KEYWORD_REGEX['fix'],
                                           [self.message])
        return {GitLabIssue(self._token, repo_name, number)
                for number, repo_name in issues}

    @property
    def will_close_issues(self) -> Set[GitLabIssue]:
        """
        Returns a set of GitLabIssue objects which would be closed as stated in
        this commit message.
        """
        issues =  self.get_keywords_issues(GITLAB_KEYWORD_REGEX['close'],
                                           [self.message])
        return {GitLabIssue(self._token, repo_name, number)
                for number, repo_name in issues}

    @property
    def will_resolve_issues(self) -> Set[GitLabIssue]:
        """
        Returns a set of GitLabIssue objects which would be resolved as stated
        in this commit message.
        """
        issues =  self.get_keywords_issues(GITLAB_KEYWORD_REGEX['resolve'],
                                           [self.message])
        return {GitLabIssue(self._token, repo_name, number)
                for number, repo_name in issues}



"""
This package contains the JIRA implementations of the interfaces in
server.git.Interfaces.
"""
from urllib.parse import urljoin
import logging
import os

from oauthlib.oauth1 import SIGNATURE_RSA
from requests_oauthlib import OAuth1

from IGitt.Interfaces import get
from IGitt.Interfaces import Token
from IGitt.Utils import CachedDataMixin
from tests import PRIVATE_KEY


JIRA_INSTANCE_URL = os.environ.get('JIRA_INSTANCE_URL',
                                   'https://jira.atlassian.com')
BASE_URL = urljoin(JIRA_INSTANCE_URL, '/rest/api/2')
JIRA_RSA_PRIVATE_KEY_PATH = os.environ.get('JIRA_RSA_PRIVATE_KEY_PATH')
JIRA_RSA_PRIVATE_KEY = PRIVATE_KEY
try:
    JIRA_RSA_PRIVATE_KEY = open(JIRA_RSA_PRIVATE_KEY_PATH, 'r').read().strip()
except (FileNotFoundError, TypeError):
    logging.warning('JIRA REST APIs work only with key signing, please '
                    'include the correct path to your registered RSA private '
                    'key.')


class JiraMixin(CachedDataMixin):
    """
    Base object for all things on Jira.
    """

    def _get_data(self):
        return get(self._token, self.url)

    @staticmethod
    def absolute_url(url):
        """
        Builds an absolute URL from the base URL and specified url.
        """
        return BASE_URL + url

    @property
    def hoster(self):
        """
        Returns `jira`.
        """
        return 'jira'

    @property
    def url(self):
        """
        Returns JIRA API url.
        """
        return self.absolute_url(self._url)

    @property
    def web_url(self):
        """
        Returns the web link for the corresponding JIRA object.
        """
        raise NotImplementedError

    def __repr__(self):  # dont cover
        return '<{} object(url={}) at {}>'.format(self.__class__.__name__,
                                                  self.url,
                                                  hex(id(self)))


class JiraOAuth1Token(Token):
    """
    Object representation of JIRA OAuth v1.0 token.
    """
    def __init__(self, client_key, key, secret):
        self.client_key = client_key
        self.key = key
        self.secret = secret

    @property
    def headers(self):
        return {}

    @property
    def parameter(self):
        return {}

    @property
    def value(self):
        return {'client_key': self.client_key,
                'oauth_token': self.key,
                'oauth_token_secret': self.secret}

    @property
    def auth(self):
        return OAuth1(self.client_key,
                      rsa_key=JIRA_RSA_PRIVATE_KEY,
                      resource_owner_key=self.key,
                      resource_owner_secret=self.secret,
                      signature_method=SIGNATURE_RSA,
                      signature_type='auth_header')



"""
This module contains the Actions classes representing the actions which
could be taken on webhooks.
"""
from enum import Enum


class MergeRequestActions(Enum):
    """
    Merge Request related actions.
    """
    # When a new merge request is created.
    OPENED = 1
    # When an existing merge request is closed.
    CLOSED = 2
    # When an existing merge request is reopened.
    REOPENED = 3
    # When someone comments on the merge request.
    COMMENTED = 4
    # When a merge request gets assigned or reassigned, labels are added or
    # removed or a milestone is changed
    ATTRIBUTES_CHANGED = 5
    # When someone pushes to an existing merge request.
    SYNCHRONIZED = 6
    # When an existing merge request is merged.
    MERGED = 7
    # When a label is added to the merge request.
    LABELED = 8
    # When a label is removed from the merge request.
    UNLABELED = 9


class IssueActions(Enum):
    """
    Issue related reactions.
    """
    # When a new issue is opened.
    OPENED = 1
    # When an existing issue is closed.
    CLOSED = 2
    # When an existing issue is reopened.
    REOPENED = 3
    # When someone comments on an issue.
    COMMENTED = 4
    # When an issue gets updated, or the linked
    # milestone is changed
    ATTRIBUTES_CHANGED = 5
    # When a new label is added to the issue
    LABELED = 6
    # When a label is removed from an issue
    UNLABELED = 7
    # When the assignees on an issue are changed.
    ASSIGNEES_CHANGED = 8

class PipelineActions(Enum):
    """
    Pipeline and Commit status related actions.
    """
    # When the status of a pipeline is updated, also depicts the change of a
    # commit status
    UPDATED = 1


class InstallationActions(Enum):
    """
    Installation and Integration related actions.
    """
    # When an app is installed.
    CREATED = 1
    # When an app is uninstalled.
    DELETED = 2
    # When a repository/repositories are added to an installation.
    REPOSITORIES_ADDED = 3
    # When a repository/repositories are removed from an installation.
    REPOSITORIES_REMOVED = 4



"""
This module contains the GitHubInstallation class which provides
properties and actions related to GitHub App installations.
"""
from functools import lru_cache
from typing import List
from typing import Set

from IGitt.GitHub import GitHubMixin
from IGitt.GitHub import GitHubInstallationToken
from IGitt.GitHub.GitHubRepository import GitHubRepository
from IGitt.Interfaces import get
from IGitt.Interfaces.Installation import Installation


class GitHubInstallation(GitHubMixin, Installation):
    """
    Represents a GitHub App installation.
    """
    def __init__(self, token: GitHubInstallationToken, installation_id: int):
        """
        Creates a GitHubInstallation object with given credentials.
        """
        self._id = installation_id
        # jwt token is used to retrieve all properties
        self._token = token.jwt
        # installation token is used only for the api
        self._api_token = token
        self._url = '/app/installations/{}'.format(self._id)

    @property
    def identifier(self) -> int:
        """
        Returns the identifier of the installation.
        """
        return self._id

    @property
    def app_id(self) -> int:
        """
        The application which was installed.
        """
        return self.data['app_id']

    @property
    def webhooks(self) -> List[str]:
        """
        Returns the list of webhooks registered by installing the application.
        """
        return self.data['events']

    @property
    def permissions(self):
        """
        The permissions granted to this installation.
        """
        return self.data['permissions']

    @property
    @lru_cache(None)
    def repositories(self) -> Set[GitHubRepository]:
        """
        Returns the set of repositories this installation has access to.
        """
        data = get(self._api_token,
                   self.absolute_url('/installation/repositories'))
        return {GitHubRepository.from_data(repo,
                                           self._api_token,
                                           repo['id'])
                for repo in data['repositories']}



"""
Contains a representation of GitHub users.
"""
from typing import Optional
from typing import Union

from IGitt import ElementDoesntExistError
from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import GitLabOAuthToken
from IGitt.GitLab import GitLabPrivateToken
from IGitt.Interfaces.User import User
from IGitt.Interfaces import get


class GitLabUser(GitLabMixin, User):
    """
    A GitLab user, e.g. sils :)
    """

    def __init__(self,
                 token: Union[GitLabPrivateToken, GitLabOAuthToken],
                 identifier: Optional[Union[str, int]]=None):
        """
        Creates a new user.

        :param token: The oauth token object
        :param identifier: Pass None if it's you! Otherwise the id, integer.
        """
        self._token = token
        self._url = '/user'
        self._id = identifier
        if identifier:
            if isinstance(identifier, str):
                params =  {'username': identifier}
                resp = get(self._token, self.absolute_url('/users'), params)
                if resp:
                    self._id = resp[0]['id']
                    self._url = '/users/' + str(resp[0]['id'])
                else:
                    raise ElementDoesntExistError('This username does not '
                                                  'exist.')
            else:
                self._url = '/users/' + str(identifier)


    @property
    def username(self) -> str:
        """
        Retrieves the login for the user.
        """
        return self.data['username']

    @property
    def identifier(self) -> int:
        """
        Gets a unique id for the user that never changes.
        """
        return self._id or self.data['id']

    def installed_repositories(self, installation_id: int):
        """
        List repositories that are accessible to the authenticated user for an
        installation.

        GitLab doesn't support building installations yet.
        """
        raise NotImplementedError

    def get_installations(self, jwt):
        """
        Gets the installations this user has access to.

        GitLab doesn't support building installations yet.
        """
        raise NotImplementedError



"""
This contains the Milestone implementation for GitLab
"""
import re
from datetime import datetime
from urllib.parse import quote_plus

from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import get
from IGitt.Interfaces import put
from IGitt.Interfaces import post
from IGitt.Interfaces import delete
from IGitt.GitLab import GitLabOAuthToken, GitLabPrivateToken
from IGitt.Interfaces.Milestone import Milestone
from IGitt.GitLab.GitLabIssue import GitLabIssue
from IGitt.GitLab.GitLabMergeRequest import GitLabMergeRequest
from IGitt.Interfaces import MilestoneStates
from IGitt.GitLab.GitLabRepository import GitLabRepository

state_translation_dict = {
    'active': MilestoneStates.OPEN,
    'closed': MilestoneStates.CLOSED,
}


class GitLabProjectMilestone(GitLabMixin, Milestone):
    """
    This class represents a project level milestone on GitLab.
    This class does not support group level milestones.
    """

    def __init__(self, token: (GitLabOAuthToken, GitLabPrivateToken), project,
                 number: int):
        """
        Creates a new GitLabProjectMilestone with the given credentials.

        :param token: A Token object to be used for authentication.
        :param project: The full name of the project. Including the owner
                        e.g. ``owner/project``.
        :param number: The unique milestone identification number (id).
                        Not The clear text number given on the Web UI
                        See also https://docs.gitlab.com/ee/api/README.html#id-vs-iid
        :raises RuntimeError: If something goes wrong (network, auth, ...)
        """
        self._token = token
        self._project = project
        self._id = number
        self._url = '/projects/{project}/milestones/{milestone_id}'.format(
            project=quote_plus(project), milestone_id=number)

    @staticmethod
    def create(token: (GitLabOAuthToken, GitLabPrivateToken),
               project,
               title: str,
               description: str = None,
               due_date=None,
               start_date=None):
        """
        Create a new milestone with given title and body.

        >>> from os import environ
        >>> milestone = GitLabProjectMilestone.create(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test',
        ...     'test milestone title',
        ...     'sample description'
        ... )
        >>> milestone.state
        'active'

        Delete the milestone to avoid filling the test repo with milestones.

        >>> milestone.close()

        :return: GitLabProjectMilestone object of the newly created milestone.
        """
        url = '/projects/{project}/milestones'.format(
            project=quote_plus(project))
        if due_date is not None:
            due_date = datetime.strftime(due_date, '%Y-%m-%d')
        if start_date is not None:
            start_date = datetime.strftime(start_date, '%Y-%m-%d')
        milestone = post(
            token, GitLabProjectMilestone.absolute_url(url), {
                'title': title,
                'description': description,
                'due_date': due_date,
                'start_date': start_date
            })

        return GitLabProjectMilestone.from_data(milestone, token, project,
                                                milestone['id'])

    @property
    def number(self) -> int:
        """
        Returns the milestone "number" or id.
        """
        return self._id

    @property
    def title(self) -> str:
        """
        Retrieves the title of the milestone.
        """
        return self.data['title']

    @title.setter
    def title(self, new_title):
        """
        Sets the title of the milestone.

        :param new_title: The new title.
        """
        self.data = put(self._token, self.url, {'title': new_title})

    @property
    def description(self) -> str:
        """
        Retrieves the main description of the milestone.
        """
        return self.data['description']

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the milestone

        :param new_description: The new description .
        """
        self.data = put(self._token, self.url,
                        {'description': new_description})

    @property
    def state(self) -> MilestoneStates:
        """
        Get's the state of the milestone.

        :return: Either MilestoneStates.OPEN or MilestoneStates.CLOSED.
        """

        return state_translation_dict[self.data['state']]

    def close(self):
        """
        Closes the milestone.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = put(self._token, self.url, {'state_event': 'close'})

    def reopen(self):
        """
        Reopens the milestone.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = put(self._token, self.url, {'state_event': 'activate'})

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was created.
        """
        return datetime.strptime(self.data['created_at'],
                                 '%Y-%m-%dT%H:%M:%S.%fZ')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was updated the last time.
        """
        return datetime.strptime(self.data['updated_at'],
                                 '%Y-%m-%dT%H:%M:%S.%fZ')

    @property
    def start_date(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was started.
        """

        return datetime.strptime(
            self.data['start_date'],
            '%Y-%m-%d') if self.data['start_date'] else None

    @start_date.setter
    def start_date(self, new_date: datetime):
        """
        Sets the start date of the milestone.

        :param new_date: The new start date.
        """
        self.data = put(
            self._token, self.url, {
                'start_date':
                datetime.strftime(new_date, '%Y-%m-%d') if new_date else None,
            })

    @property
    def due_date(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone is due.
        """

        return datetime.strptime(self.data['due_date'],
                                 '%Y-%m-%d') if self.data['due_date'] else None

    @due_date.setter
    def due_date(self, new_date: datetime):
        """
        Sets the due date of the milestone.

        :param new_date: The new due date.
        """

        self.data = put(
            self._token, self.url, {
                'due_date':
                datetime.strftime(new_date, '%Y-%m-%d') if new_date else None
            })

    @staticmethod
    def extract_repo_full_name(web_url):
        """
        Extracts the repository name from the web_url of the issue
        """
        return re.sub(r'https?://gitlab\.com/|/issues/\d', '', web_url)

    @property
    def issues(self) -> set:
        """
        Retrieves a set of issue objects that are assigned to this milestone.
        """
        return {
            GitLabIssue.from_data(res, self._token,
                                  self.extract_repo_full_name(res['web_url']),
                                  res['iid'])
            for res in get(self._token, self.url + '/issues')
        }

    @property
    def merge_requests(self) -> set:
        """
        Retrieves a set of merge request objects that are assigned to this
        milestone.
        """
        return {
            GitLabMergeRequest.from_data(res, self._token,
                                         self.extract_repo_full_name(
                                             res['web_url']), res['iid'])
            for res in get(self._token, self.url + '/merge_requests')
        }

    @property
    def project(self) -> GitLabRepository:
        """
        Returns the repository this milestone is linked with.
        """
        return GitLabRepository(self._token, self._project)

    def delete(self):
        """
        Deletes the milestone.
        This is not possible with GitLab api v4.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        delete(self._token, self.url)



"""
Provides useful stuff, generally!
"""
from datetime import datetime
from collections import OrderedDict
from typing import Callable
from typing import Optional
import json

from pytz import timezone


class LimitedSizeDict(OrderedDict):
    """
    LimitedSizeDict pops items from the first if the size of dictionary exceeds
    the specified limit.
    """
    def __init__(self, *args, **kwargs):
        self.size_limit = kwargs.pop('size_limit', None)
        super(LimitedSizeDict, self).__init__(*args, **kwargs)
        self._check_size_limit()

    def __setitem__(self, *args, **kwargs):
        super(LimitedSizeDict, self).__setitem__(*args, **kwargs)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while self.size_limit < len(self):
                self.popitem(last=False)


class Cache:
    """
    A class to manage cache with IGitt and any other external application.

    The cache mechanism should be able to process raw JSON data. The response
    data from external requests is stored in the cache with the API URL of the
    associated IGitt object as the key and the data is stored along with its
    entity tag header, which is used alongside `If-None-Match` header for
    further queries using conditional requests. When an incoming webhook is
    received, the timestamp of reception is cached and any queries later on the
    same URL use the `If-Modified-Since` HTTP Header and the reception time
    using conditional requests, since the ETag Header is no longer valid. Note
    that conditional requests do not add up to rate limits on APIs.

    To use IGitt's caching mechanism for external request management, simply
    add the following code to your application before using IGitt.

    >>> from IGitt.Utils import Cache
    >>> Cache.use(read_from, write_to)

    If not provided, IGitt uses a default in-memory cache. For further details
    follow the specific method documentation below.
    """
    __mem_store = LimitedSizeDict(size_limit=10 ** 6)  # a million entries
    _get = __mem_store.__getitem__
    _set = __mem_store.__setitem__

    @classmethod
    def use(cls, read_from: Callable, write_to: Callable):
        """
        Connects the cache read, write functions to Cache class.

        :param read_from:
            The method to be called to fetch data from cache. It should be able
            to receive only one parameter, key, which is used to identify an
            entry uniquely in the cache.
        :param write_to:
            The method to be called to write data to cache. It should be able
            to receive two parameters, key (used to identify the entry in
            cache) and the item to be stored in cache, in the specified
            respective order.
        """
        cls._get = read_from
        cls._set = write_to

    @classmethod
    def validate(cls, item: dict) -> dict:
        """
        Checks if the given item has valid data and adds missing fields with
        default values. Expected fields are as follows.

        :type fromWebhook:  bool
        :type data:         object
        :type links:        dict
        :type lastUpdated:  str (formatted as '%a, %d %m %Y %H:%M:%S %Z')
        :type entityTag:    str or None

        :return:    The item dictionary after validation without any missing
                    fields. Also removes any additional unrelated fields from
                    the given dictionary.
        :raises:    TypeError, if the field does not match the expected type.
                    ValueError, if the type is correct, but the value is
                    invalid.
        """
        if 'data' not in item:
            item['data'] = {}

        if 'links' not in item:
            item['links'] = {}
        elif not isinstance(item['links'], dict):
            raise TypeError("'links' field should be a dictionary, not {}"
                            ''.format(type(item['links'])))

        if 'lastFetched' not in item:
            item['lastFetched'] = datetime.now(
                timezone('GMT')).strftime('%a, %d %m %Y %H:%M:%S %Z')
        else:
            # check if the datetime format is correct and raises an exception
            # if it is invalid. TypeError, if item['lastFetched'] is not a
            # string and ValueError, if the format doesn't match the expected.
            datetime.strptime(item['lastFetched'],
                              '%a, %d %m %Y %H:%M:%S %Z')

        if 'entityTag' not in item:
            item['entityTag'] = None
        elif (not isinstance(item['entityTag'], str) and
              item['entityTag'] != None):
            raise TypeError(
                "'entityTag' field should either be a string or None, not {}"
                ''.format(type(item['entityTag'])))

        if 'fromWebhook' not in item:
            item['fromWebhook'] = False
        elif not isinstance(item['fromWebhook'], bool):
            raise TypeError("'fromWebhook' field should be a bool, not {}"
                            ''.format(type(item['fromWebhook'])))

        # drop any other extra fields in the dictionary
        fields = {'fromWebhook', 'entityTag', 'lastFetched', 'links', 'data'}
        item = {k: v for k, v in item.items() if k in fields}

        return item

    @classmethod
    def get(cls, key) -> Optional[dict]:
        """
        Retrieves the entry from cache if present, otherwise None.
        """
        try:
            return cls.validate(json.loads(cls._get(key)))
        except (KeyError, TypeError):
            return None

    @classmethod
    def set(cls, key, item):
        """
        Stores the entry in cache.
        """
        item = cls.validate(item)
        cls._set(key, json.dumps(item))

    @classmethod
    def update(cls, key, new_value):
        """
        Updates the existing entry with new data, if present, otherwise creates
        a new entry in cache.
        """
        cls.set(key, {**(cls.get(key) or {}), **new_value})


class PossiblyIncompleteDict:
    """
    A dict kind of thing (only supporting item getting) that, if an item isn't
    available, gets fresh data from a refresh function.
    """

    def __init__(self, data: dict, refresh) -> None:
        self.may_need_refresh = True
        self._data = self._del_nul(data)
        self._refresh = refresh

    @staticmethod
    def _del_nul(elem):
        """
        elegantly tries to remove invalid \x00 chars from
        strings, strings in lists, strings in dicts.
        """
        if isinstance(elem, str):
            return elem.replace(chr(0), '')

        elif isinstance(elem, dict):
            return {key: PossiblyIncompleteDict._del_nul(value)
                    for key, value in elem.items()}

        elif isinstance(elem, list):
            return [PossiblyIncompleteDict._del_nul(item)
                    for item in elem]

        return elem

    def __getitem__(self, item):
        if item in self._data:
            return self._data[item]

        self.maybe_refresh()
        return self._data[item]

    def __setitem__(self, key, item):
        self._data[key] = item

    def __contains__(self, item):
        """
        Needed for determining if an item is in the possibly incomplete self.
        """
        return item in self._data

    def update(self, value: dict):
        """
        Updates the dict with provided dict.
        """
        self._data.update(self._del_nul(value))

    def maybe_refresh(self):
        """
        Refresh if it may need a refresh.
        """
        if self.may_need_refresh:
            self.refresh()

    def refresh(self):
        """
        Refreshes data unconditionally.
        """
        self._data = self._del_nul(self._refresh())
        self.may_need_refresh = False

    def get(self):
        """
        Returns the stored data.
        """
        return self._data


class CachedDataMixin:
    """
    You provide:

    - self._get_data for getting your data

    You can also create an IGitt instance with your own data using from_data
    classmethod.
    """
    default_data = {}  # type: dict

    @classmethod  # Ignore PyLintBear
    def from_data(cls, data: Optional[dict]=None, *args, **kwargs):
        """
        Returns an instance created from the provided data. No further requests
        are made.

        :raises TypeError:
            When the args provided are insufficient to call __init__.
        """
        instance = cls(*args, **kwargs)
        instance.data = data or {}

        return instance

    def _get_data(self):
        """
        Retrieves the data for the object.
        """
        raise NotImplementedError

    def refresh(self):  # dont cover
        """
        Refreshes all the data from the hoster!
        """
        if not getattr(self, '_data', None):
            self._data = PossiblyIncompleteDict(
                self.default_data, self._get_data)

        self._data.refresh()

    @property
    def data(self):
        """
        Retrieves the data, if needed from the network.
        """
        if not getattr(self, '_data', None):
            self._data = PossiblyIncompleteDict(
                self.default_data, self._get_data)

        return self._data

    @data.setter
    def data(self, value):
        """
        Setter for the data, use it to override, refresh, ...
        """
        self._data = PossiblyIncompleteDict(value, self._get_data)


def eliminate_none(data):
    """
    Remove None values from dict
    """
    return dict((k, v) for k, v in data.items() if v is not None)



"""
This contains the Milestone implementation for GitHub
"""
from datetime import datetime

from IGitt.GitHub import GitHubMixin
from IGitt.Interfaces.Milestone import Milestone
from IGitt.Interfaces import MilestoneStates
from IGitt.GitHub import GitHubToken
from IGitt.Interfaces import post
from IGitt.Interfaces import patch
from IGitt.Interfaces import delete
from IGitt.Interfaces import get
from IGitt.GitHub.GitHubIssue import GitHubIssue
from IGitt.GitHub.GitHubMergeRequest import GitHubMergeRequest
from IGitt.GitHub.GitHubRepository import GitHubRepository


class GitHubMilestone(GitHubMixin, Milestone):
    """
    This class represents a milestone on GitHub.
    """

    def __init__(self, token: GitHubToken, repository: str, number: int):
        """
        Creates a new GitHubMilestone object with the given credentials.

        :param token: A Token object to be used for authentication.
        :param repository: The full name of the repository
                            eg. 'sils/repository'.
        :param number: The milestones number.
        :raises RuntimeError: If something goes wrong (network, auth, ...)
        """
        self._token = token
        self._repository = repository
        self._number = number
        self._url = '/repos/{repository}/milestones/{milestone_number}'\
            .format(repository=repository, milestone_number=number)
        self._issues_url = GitHubMixin.absolute_url(
            '/repos/{repository}/issues'.format(repository=self._repository))

    @staticmethod
    def create(token: GitHubToken,
               repository: str,
               title: str,
               state: MilestoneStates = MilestoneStates.OPEN,
               description: str = None,
               due_on: datetime = None):
        """
        Create a new milestone with given title
        :return: GitHubMilestone object of the newly created milestone.
        """
        url = '/repos/{repository}/milestones'.format(repository=repository)
        if due_on != None:
            due_on = datetime.strftime(due_on, '%Y-%m-%dT%H:%M:%SZ')
        milestone = post(
            token, GitHubMilestone.absolute_url(url), {
                'title': title,
                'state': str(state),
                'description': description,
                'due_on': due_on
            })
        return GitHubMilestone.from_data(milestone, token, repository,
                                         milestone['number'])

    @property
    def number(self) -> int:
        """
        Returns the milestone "number" or id.
        """
        return self._number

    @property
    def title(self) -> str:
        """
        Retrieves the title of the milestone.
        """
        return self.data['title']

    @title.setter
    def title(self, new_title):
        """
        Sets the title of the milestone.

        :param new_title: The new title.
        """
        self.data = patch(self._token, self.url, {'title': new_title})

    @property
    def description(self) -> str:
        """
        Retrieves the main description of the milestone.
        """
        return self.data['description']

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the milestone

        :param new_description: The new description .
        """
        self.data = patch(self._token, self.url,
                          {'description': new_description})

    @property
    def state(self) -> MilestoneStates:
        """
        Get's the state of the milestone.

        :return: Either MilestoneStates.OPEN or MilestoneStates.CLOSED.
        """
        return MilestoneStates[self.data['state'].upper()]

    def close(self):
        """
        Closes the milestone.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = patch(self._token, self.url, {'state': 'closed'})

    def reopen(self):
        """
        Reopens the milestone.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = patch(self._token, self.url, {'state': 'open'})

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was created.
        """
        return datetime.strptime(self.data['created_at'], '%Y-%m-%dT%H:%M:%SZ')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was updated the last time.
        """
        return datetime.strptime(self.data['updated_at'], '%Y-%m-%dT%H:%M:%SZ')

    @property
    def due_date(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone is due.
        """

        return datetime.strptime(
            self.data['due_on'],
            '%Y-%m-%dT%H:%M:%SZ') if self.data['due_on'] else None

    @due_date.setter
    def due_date(self, new_date: datetime):
        """
        Sets the due date of the milestone.
        It is not possible to set the time. GitHub will always set the time on
        the due date to 07:00:00

        :param new_date: The new due date.
        """
        self.data = patch(
            self._token, self.url, {
                'due_on':
                datetime.strftime(new_date, '%Y-%m-%dT%H:%M:%SZ')
                if new_date else None
            })

    def delete(self):
        """
        Deletes the milestone.
        This is not possible with GitLab api v4.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        delete(self._token, self.url)

    @property
    def issues(self) -> set:
        """
        Retrieves a set of issue objects that are assigned to this milestone.
        """
        return {
            GitHubIssue.from_data(res, self._token, self._repository,
                                  res['number'])
            for res in get(self._token, self._issues_url,
                           {'milestone': self._number})
            if 'pull_request' not in res
        }

    @property
    def merge_requests(self) -> set:
        """
        Retrieves a set of merge_request
        objects that are assigned to this milestone.
        """
        return {
            GitHubMergeRequest.from_data(res, self._token, self._repository,
                                         res['number'])
            for res in get(self._token, self._issues_url,
                           {'milestone': self._number})
            if 'pull_request' in res
        }

    @property
    def project(self) -> GitHubRepository:
        """
        Returns the repository this milestone is linked with.
        """
        return GitHubRepository(self._token, self._repository)

    @property
    def start_date(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was started.
        The start_date does not exist in GitHub.
        """
        return None



"""
Contains an object representation of a reaction on GitHub.
"""
from functools import lru_cache
from typing import Union

from IGitt.GitHub import GitHubMixin
from IGitt.GitHub import GitHubToken
from IGitt.GitHub.GitHubUser import GitHubUser
from IGitt.Interfaces import get
from IGitt.Interfaces.Comment import Comment
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces.MergeRequest import MergeRequest
from IGitt.Interfaces.Reaction import Reaction


PREVIEW_HEADER = {'Accept': 'application/vnd.github.squirrel-girl-preview'}


class GitHubReaction(GitHubMixin, Reaction):
    """
    A GitHub reaction, e.g. heart.
    """
    @lru_cache(None)
    def _get_data(self):
        # Note: A GitHub reaction cannot be retrieved using a GET request, it
        # has to retrieved as a list and filtered for the match.
        if not getattr(self, '_list', None):
            setattr(self, '_list', get(self._token, self.url,
                                       headers=PREVIEW_HEADER))
        try:
            return list(filter(lambda x: x['id'] == self._identifier,
                               getattr(self, '_list')))[0]
        except IndexError:
            raise RuntimeError({
                'message': 'Not Found',
                'documentation_url': 'https://developer.github.com/v3'}, 404)

    def __init__(self,
                 token: GitHubToken,
                 related: Union[Issue, MergeRequest, Comment],
                 identifier: int):
        """
        Creates a GitHubReaction instance.

        :param token:       The authentication token.
        :param related:     The object this reaction was added on,
                            i.e. an Issue, MergeRequest, Comment, etc.
        :param identifier:  The unique identifier of the reaction.
        """
        self._token = token
        self._related = related
        self._url = '{}/reactions'.format(getattr(related, '_url'))
        self._identifier = identifier

    @property
    def name(self) -> str:
        """
        Retrieves the name of the reaction.
        """
        return self.data['content']

    @property
    def user(self) -> GitHubUser:
        """
        Retrieves the user who reacted with this reaction.
        """
        user = self.data['user']
        return GitHubUser.from_data(user, self._token, user['login'])

    def __hash__(self):
        """
        Unique hash for this thing.
        """
        return hash(self.url + str(self._identifier))



"""
Contains the GitLab Team implementation.
"""
from typing import Set

from IGitt.GitLab import GitLabMixin
from IGitt.GitLab.GitLabUser import GitLabUser
from IGitt.Interfaces import get

class GitLabTeam(GitLabMixin):
    """
    Represents a Team on GitLab.
    """

    def __init__(self, token, group_id):
        """
        :param group_id: The id of the group.
        """
        self._token = token
        self._id = group_id
        self._url = '/groups/{group_id}'.format(group_id=group_id)

    @property
    def name(self) -> str:
        """
        Name of the team.
        """
        return self.data['name']

    @property
    def id(self) -> int:
        """
        Retrieves the id of the team.
        """
        return self._id

    @property
    def description(self) -> str:
        """
        Returns the description of this team.
        """
        return self.data['description']

    @property
    def members(self) -> Set[GitLabUser]:
        """
        Returns the user handles of all members of this team.
        """
        return {
            GitLabUser.from_data(user, self._token, user['id'])
            for user in get(
                self._token, self.url + '/members'
            )
        }

    def is_member(self, username):
        """
        Checks if given username is member of this team.
        """
        usernames = [user.username for user in self.members]
        return True if username in usernames else False

    @property
    def get_organization(self):
        """
        Returns parent team.
        """
        if self.data['parent_id']:
            return GitLabTeam(self._token, self.data['parent_id'])
        return None



"""
Contains a class that represents a request to merge something into some git
branch.
"""
from datetime import datetime
from typing import List
from typing import Set

from IGitt.Interfaces import MergeRequestStates
from IGitt.Interfaces.Commit import Commit
from IGitt.Interfaces.CommitStatus import Status
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces.User import User
from IGitt.Interfaces.Milestone import Milestone


class MergeRequest(Issue):
    """
    A request to merge something into the main codebase. Can be a patch in a
    mail or a pull request on GitHub.
    """

    def close(self):
        """
        Closes the merge request.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def reopen(self):
        """
        Reopens the merge request.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @property
    def state(self) -> MergeRequestStates:
        """
        Get's the state of the merge request.

        :return: A MergeRequestStates object indicating current state.
        """
        raise NotImplementedError

    @property
    def base(self) -> Commit:
        """
        Retrieves the base commit of the merge request, i.e. the one it should
        be merged into.

        :return: A Commit object.
        """
        raise NotImplementedError

    @property
    def base_branch_name(self) -> str:
        """
        Retrieves the base branch name of the merge request, i.e. the one it
        should be merged into.

        :return: A string.
        """
        raise NotImplementedError

    @property
    def head(self) -> Commit:
        """
        Retrieves the head commit of the merge request, i.e. the one which
        would be merged.

        :return: A Commit object.
        """
        raise NotImplementedError

    @property
    def head_branch_name(self) -> str:
        """
        Retrieves the head branch name of the merge request, i.e. the one that
        will be merged.

        :return: A string.
        """
        raise NotImplementedError

    @property
    def commits(self) -> List[Commit]:
        """
        Retrieves all commits that are contained in this request.

        :return: A list of Commits.
        """
        raise NotImplementedError

    @property
    def repository(self):
        """
        Retrieves the repository where this PR is opened at.

        :return: A Repository object.
        """
        raise NotImplementedError

    @property
    def target_repository(self):
        """
        Retrieves the repository where this PR is opened at. An alias to
        ``self.repository`` property.

        :return: A Repository object.
        """
        return self.repository

    @property
    def source_repository(self):
        """
        Retrieves the repository where this PR's head branch is located at.

        :return: A Repository object.
        """
        raise NotImplementedError

    @property
    def affected_files(self):
        """
        Retrieves the affected files.

        :return: A set of filenames relative to repo root.
        """
        raise NotImplementedError

    @property
    def diffstat(self):
        """
        Gets additions and deletions of a merge request.

        :return: An (additions, deletions) tuple.
        """
        raise NotImplementedError

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the merge request was created.
        """
        raise NotImplementedError

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the merge request was updated the last
        time.
        """
        raise NotImplementedError

    @property
    def number(self) -> int:
        """
        Returns the MR "number" or id.
        """
        raise NotImplementedError

    @property
    def closes_issues(self) -> Set[Issue]:
        """
        Returns a set of Issue objects which would be closed upon merging this
        pull request.
        """
        raise NotImplementedError

    @property
    def tests_passed(self) -> bool:
        """
        Returns True if all commits of the merge request have a success state.
        If you wish to only get the head state, use mr.head.combined_status.
        """
        statuses = set(map(lambda commit: commit.combined_status,
                           self.commits))
        if Status.PENDING in statuses or Status.FAILED in statuses:
            return False
        return True

    @property
    def mentioned_issues(self) -> Set[Issue]:
        """
        Returns a set of Issue objects which are related to the pull request.
        """
        raise NotImplementedError

    @property
    def author(self) -> User:
        """
        Returns the author of the MR wrapped in a User object.
        """
        raise NotImplementedError

    def merge(self, message: str=None, sha: str=None,
              should_remove_source_branch: bool=False,
              _github_merge_method: str=None,
              _gitlab_merge_when_pipeline_succeeds: bool=False):
        """
        Merges the merge request.

        :param message:                     The commit message.
        :param sha:                         The commit sha that the HEAD must
                                            match in order to merge.
        :param should_remove_source_branch: Whether the source branch should be
                                            removed upon a successful merge.
        :param _github_merge_method:        On GitHub, the merge method to use
                                            when merging the MR. Can be one of
                                            `merge`, `squash` or `rebase`.
        :param _gitlab_wait_for_pipeline:   On GitLab, whether the MR should be
                                            merged immediately after the
                                            pipeline succeeds.
        :raises RuntimeError:        If something goes wrong (network, auth...).
        :raises NotImplementedError: If an unused parameter is passed.
        """
        raise NotImplementedError

    @property
    def milestone(self) -> Milestone:
        """
        Retrieves the milestone.
        """
        raise NotImplementedError

    @milestone.setter
    def milestone(self, new_milestone) -> Milestone:
        """
        Setter for the Milestone.
        """
        raise NotImplementedError

    @property
    def mergeable(self) -> bool:
        """
        Returns true if there is no merge conflict.
        """
        raise NotImplementedError
